/*******************************************************************************

 Edited version of the Dome Simulator
 Copyright(c) 2014 Jasem Mutlaq. All rights reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License version 2 as published by the Free Software Foundation.
 .
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.
 .
 You should have received a copy of the GNU Library General Public License
 along with this library; see the file COPYING.LIB.  If not, write to
 the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 Boston, MA 02110-1301, USA.
*******************************************************************************/


/*
 * Modified version of the rolloff roof simulator.
 * Uses a simple text string protocol to send messages to an Arduino. The Arduino's USB connection is set
 * by default to 38400 baud. The Arduino code determines which pins open/close a relay to start/stop the
 * roof motor, and read state of switches indicating if roof opened or closed.
 */
#include "rolloffino.h"
#include "indicom.h" 
#include "termios.h"

#include <cmath>
#include <cstring>
#include <ctime>
#include <memory>

#define ROLLOFF_DURATION 15                 // Seconds until Roof is fully opened or closed
#define ROR_D_PRESS      1000                // Milliseconds after issuing command before expecting response

// Read only
#define ROOF_OPENED_SWITCH "OPENED"
#define ROOF_CLOSED_SWITCH "CLOSED"
#define ROOF_LOCKED_SWITCH "LOCKED"
#define ROOF_AUX_SWITCH    "AUXSTATE"

// Write only
#define ROOF_OPEN_RELAY     "OPEN"
#define ROOF_CLOSE_RELAY    "CLOSE"
#define ROOF_ABORT_RELAY    "ABORT"
#define ROOF_AUX_RELAY      "AUXSET"

// Arduino controller interface limits
#define MAXINOCMD        15          // Command buffer
#define MAXINOTARGET     15          // Target buffer
#define MAXINOVAL        127         // Value bufffer, sized to contain NAK error strings
#define MAXINOLINE       63          // Sized to contain outgoing command requests
#define MAXINOBUF        255         // Sized for maximum overall input / output
#define MAXINOERR        255         // System call error message buffer
#define MAXINOWAIT       2  // seconds

// We declare an auto pointer to RollOffIno.
std::unique_ptr<RollOffIno> rollOffIno(new RollOffIno());

void ISPoll(void *p);

void ISGetProperties(const char *dev)
{
    rollOffIno->ISGetProperties(dev);
}

void ISNewSwitch(const char *dev, const char *name, ISState *states, char *names[], int n)
{
    rollOffIno->ISNewSwitch(dev, name, states, names, n);
}

void ISNewText(const char *dev, const char *name, char *texts[], char *names[], int n)
{
    rollOffIno->ISNewText(dev, name, texts, names, n);
}

void ISNewNumber(const char *dev, const char *name, double values[], char *names[], int n)
{
    rollOffIno->ISNewNumber(dev, name, values, names, n);
}

void ISNewBLOB(const char *dev, const char *name, int sizes[], int blobsizes[], char *blobs[], char *formats[],
               char *names[], int n)
{
    INDI_UNUSED(dev);
    INDI_UNUSED(name);
    INDI_UNUSED(sizes);
    INDI_UNUSED(blobsizes);
    INDI_UNUSED(blobs);
    INDI_UNUSED(formats);
    INDI_UNUSED(names);
    INDI_UNUSED(n);
}

void ISSnoopDevice(XMLEle *root)
{
    rollOffIno->ISSnoopDevice(root);
}


bool RollOffIno::ISSnoopDevice(XMLEle *root)
{
    return INDI::Dome::ISSnoopDevice(root);
}

RollOffIno::RollOffIno()
{
    SetDomeCapability(DOME_CAN_ABORT | DOME_CAN_PARK);           // Need the DOME_CAN_PARK capability for the scheduler
}

/**************************************************************************************
** INDI is asking us for our default device name. 
** Check that it matches Ekos selection menu and ParkData.xml names
***************************************************************************************/
const char *RollOffIno::getDefaultName()
{
    return (const char *)"RollOff ino";
}
/**************************************************************************************
** INDI request to init properties. Connected Define properties to Ekos
***************************************************************************************/
bool RollOffIno::initProperties()
{
    INDI::Dome::initProperties();

    IUFillSwitch(&AuxS[AUX_DISABLE], "AUX_DISABLE", "Off", ISS_ON);
    IUFillSwitch(&AuxS[AUX_ENABLE], "AUX_ENABLE", "On", ISS_OFF);
    IUFillSwitchVector(&AuxSP, AuxS, 2, getDeviceName(), "AUX", "Auxiliary", MAIN_CONTROL_TAB, IP_RW, ISR_1OFMANY, 60, IPS_IDLE);

    IUFillLight(&RoofStatusL[ROOF_STATUS_OPENED], "Opened", "", IPS_IDLE);
    IUFillLight(&RoofStatusL[ROOF_STATUS_CLOSED], "Closed", "", IPS_IDLE);
    IUFillLight(&RoofStatusL[ROOF_STATUS_MOVING], "Moving", "", IPS_IDLE);
    IUFillLight(&RoofStatusL[ROOF_STATUS_LOCKED], "Roof Lock", "", IPS_IDLE);
    IUFillLight(&RoofStatusL[ROOF_STATUS_AUXSTATE], "Roof Auxiliary", "", IPS_IDLE);
    IUFillLightVector(&RoofStatusLP, RoofStatusL, 5, getDeviceName(), "Roof Status", "", MAIN_CONTROL_TAB, IPS_BUSY);

    SetParkDataType(PARK_NONE);
    addAuxControls();         // This is for standard controls not the local auxiliary switch
    return true;
}

/************************************************************************************
 * Called from Dome, BaseDevice to establish contact with device
 ************************************************************************************/
bool RollOffIno::Handshake()
{
    bool status = false;

    if (PortFD <= 0)
        DEBUG(INDI::Logger::DBG_WARNING,"The connection port has not been established");
    else
    {
        if (!(status = initialContact()))
        {
            msSleep(1000);              // In case it is a Nano that was still resetting
            status = initialContact();
        }

        if (!status)
            DEBUG(INDI::Logger::DBG_WARNING,"Initial controller contact failed");
    }
    return status;
}

/**************************************************************************************
** Client is asking us to establish connection to the device
***************************************************************************************/
bool RollOffIno::Connect()
{
    bool status = INDI::Dome::Connect();
    return status;
}

/**************************************************************************************
** Client is asking us to terminate connection to the device
***************************************************************************************/
bool RollOffIno::Disconnect()
{
    bool status = INDI::Dome::Disconnect();
    return status;
}

/********************************************************************************************
** INDI request to update the properties because there is a change in CONNECTION status
** This function is called whenever the device is connected or disconnected.
*********************************************************************************************/
bool RollOffIno::updateProperties()
{
    INDI::Dome::updateProperties();
    if (isConnected())
    {
        defineLight(&RoofStatusLP);    // All the roof status lights
        defineSwitch(&AuxSP);          // Aux Switch,
        setupConditions();
    }
    else
    {
        deleteProperty(RoofStatusLP.name);  // Delete the roof status lights
        deleteProperty(AuxSP.name);         // Delete the Auxiliary Switch buttons
     }
    return true;
}

/********************************************************************************************
** Establish conditions on a connect.
*********************************************************************************************/
bool RollOffIno::setupConditions()
{
    updateRoofStatus();

    // If we have Dome parking data
    if (InitPark())
    {
        DEBUG(INDI::Logger::DBG_SESSION, "Dome parking data was obtained");
    }
        // If we do not have Dome parking data
    else
     {
         DEBUG(INDI::Logger::DBG_SESSION,"Dome parking data was not obtained");
     }

    Dome::DomeState curState = getDomeState();
    switch (curState)
    {
        case DOME_UNKNOWN:
            DEBUG(INDI::Logger::DBG_SESSION,"Dome state: DOME_UNKNOWN");
            break;
        case    DOME_ERROR:
            DEBUG(INDI::Logger::DBG_SESSION,"Dome state: DOME_ERROR");
            break;
        case DOME_IDLE:
            DEBUG(INDI::Logger::DBG_SESSION,"Dome state: DOME_IDLE ");
            break;
        case     DOME_MOVING:
            DEBUG(INDI::Logger::DBG_SESSION,"Dome state: DOME_MOVING");
            break;
        case     DOME_SYNCED:
            DEBUG(INDI::Logger::DBG_SESSION,"Dome state: DOME_SYNCED");
            break;
        case     DOME_PARKING:
            DEBUG(INDI::Logger::DBG_SESSION,"Dome state: DOME_PARKING");
            break;
        case    DOME_UNPARKING:
            DEBUG(INDI::Logger::DBG_SESSION,"Dome state: DOME_UNPARKING");
            break;
        case    DOME_PARKED:
            if (isParked())
            {
                DEBUG(INDI::Logger::DBG_SESSION,"Dome state: DOME_PARKED");
            }
            else
            {
                DEBUG(INDI::Logger::DBG_SESSION,"Dome state is DOME_PARKED but Dome status is unparked");
            }
            break;
        case    DOME_UNPARKED:
            if (!isParked())
            {
                DEBUG(INDI::Logger::DBG_SESSION,"Dome state: DOME_UNPARKED");
            }
            else
            {
                DEBUG(INDI::Logger::DBG_SESSION,"Dome state is DOME_UNPARKED but Dome status is parked");
            }
            break;

    }

     // If Dome:: park status different from roof status, just o/p message (the roof might be or need to be operating manually)
     // If park status is the same, and Dome:: state does not match, change the Dome:: state.
     if (isParked())
     {
         if (fullClosedLimitSwitch == ISS_OFF)
         {
                DEBUG(INDI::Logger::DBG_WARNING,"Dome indicates it is parked but roof closed switch not set");
         }
         else
         {
             // When Dome status agrees with roof status (closed), but Dome state differs, set Dome state to parked
             if (curState != DOME_PARKED)
             {
                 DEBUG(INDI::Logger::DBG_SESSION,"Setting Dome state to DOME_PARKED.");
                 setDomeState(DOME_PARKED);
             }
         }
     }
     else
     {
         if (fullOpenLimitSwitch == ISS_OFF)
         {
             DEBUG(INDI::Logger::DBG_WARNING,"Dome indicates it is unparked but roof open switch is not set");
         }
         else
             // When Dome status agrees with roof status (open), but Dome state differs, set Dome state to unparked
         {
             if (curState != DOME_UNPARKED)
             {
                 DEBUG(INDI::Logger::DBG_SESSION,"Setting Dome state to DOME_UNPARKED.");
                 setDomeState(DOME_UNPARKED);
             }
         }
     }
    return true;
}

/********************************************************************************************
** Client has changed the state of a switch, update
*********************************************************************************************/
bool RollOffIno::ISNewSwitch(const char *dev, const char *name, ISState *states, char *names[], int n)
{
    // Make sure the call is for our device
    if(dev != nullptr && strcmp(dev,getDeviceName()) == 0)
    {
        // Check if the call for our Aux switch
        if (strcmp(name, AuxSP.name) == 0)
        {
            // Find out which state is requested by the client
            const char *actionName = IUFindOnSwitchName(states, names, n);
            // If it is the same state as actionName, then we do nothing. i.e.
            // if actionName is AUX_ON and our Aux switch is already on, we return
            int currentAuxIndex = IUFindOnSwitchIndex(&AuxSP);
            if (!strcmp(actionName, AuxS[currentAuxIndex].name))
            {
                DEBUGF(INDI::Logger::DBG_SESSION, "Aux enable is already %s", AuxS[currentAuxIndex].label);
                AuxSP.s = IPS_IDLE;
                IDSetSwitch(&AuxSP, NULL);
                return true;
            }
            // Update the switch state
            IUUpdateSwitch(&AuxSP, states, names, n);
            currentAuxIndex = IUFindOnSwitchIndex(&AuxSP);
            DEBUGF(INDI::Logger::DBG_SESSION, "Roof Auxiliary is now %s", AuxS[currentAuxIndex].label);
            AuxSP.s = IPS_OK;
            IDSetSwitch(&AuxSP, nullptr);
            updateRoofStatus();
        }
    }
    return INDI::Dome::ISNewSwitch(dev, name, states, names, n);
}

void RollOffIno::updateRoofStatus()
{
    bool auxState = false;
    bool lockState = false;
    bool openState = false;
    bool closeState = false;

    if (!getFullOpenedLimitSwitch(&openState))
    {
        DEBUG(INDI::Logger::DBG_WARNING,"Roof opened switch could not be obtained from controller");
    }
    if (!getFullClosedLimitSwitch(&closeState))
    {
        DEBUG(INDI::Logger::DBG_WARNING,"Roof closed switch could not be obtained from controller");
    }
    if (!getRoofLockedSwitch(&lockState))
    {
        DEBUG(INDI::Logger::DBG_WARNING,"Roof lock switch could not be obtained from controller");
    }
    if (!getRoofAuxSwitch(&auxState))
    {
        DEBUG(INDI::Logger::DBG_WARNING,"Roof aux switch could not be obtained from controller");
    }

    RoofStatusL[ROOF_STATUS_AUXSTATE].s = IPS_IDLE;
    RoofStatusL[ROOF_STATUS_LOCKED].s = IPS_IDLE;
    RoofStatusL[ROOF_STATUS_OPENED].s = IPS_IDLE;
    RoofStatusL[ROOF_STATUS_CLOSED].s = IPS_IDLE;
    RoofStatusL[ROOF_STATUS_MOVING].s = IPS_IDLE;
    RoofStatusLP.s = IPS_IDLE;

    if (auxState)
    {
        RoofStatusL[ROOF_STATUS_AUXSTATE].s = IPS_OK;
    }

    if (lockState)
    {
        if (closeState)
        {
            RoofStatusL[ROOF_STATUS_LOCKED].s = IPS_ALERT;         // Red to indicate lock is on
            RoofStatusL[ROOF_STATUS_CLOSED].s = IPS_OK;            // Closed and locked roof status is normal
            RoofStatusLP.s = IPS_OK;
        } else
        {
            RoofStatusL[ROOF_STATUS_LOCKED].s = IPS_ALERT;         // Possible, but should not be locked and opened at the same time
            RoofStatusL[ROOF_STATUS_CLOSED].s = IPS_ALERT;
            RoofStatusLP.s = IPS_ALERT;
        }
    }
    else
    {
        if ((openState || closeState) && (roofMoving || roofOpening || roofClosing))
        {
            roofMoving = false;
            roofOpening = false;
            roofClosing = false;
        }
        if (openState && !closeState)
        {
            RoofStatusL[ROOF_STATUS_OPENED].s = IPS_OK;
            RoofStatusLP.s = IPS_OK;
        }
        else if (closeState && !openState)
        {
            RoofStatusL[ROOF_STATUS_CLOSED].s = IPS_OK;
            RoofStatusLP.s = IPS_OK;
        }
        else if (roofOpening || roofClosing)
        {
            if (roofOpening)
            {
                RoofStatusL[ROOF_STATUS_OPENED].s = IPS_BUSY;
                RoofStatusL[ROOF_STATUS_MOVING].s = IPS_BUSY;
            }
            else if (roofClosing)
            {
                RoofStatusL[ROOF_STATUS_CLOSED].s = IPS_BUSY;
                RoofStatusL[ROOF_STATUS_MOVING].s = IPS_BUSY;
            }
            RoofStatusLP.s = IPS_BUSY;
            roofMoving = true;
        }
        // Roof is stationary, neither opened or closed
        else
        {
            RoofStatusLP.s = IPS_ALERT;
        }
    }
    IDSetLight(&RoofStatusLP, nullptr);
}

/********************************************************************************************
** Each 1 second timer tick, if roof active
********************************************************************************************/
void RollOffIno::TimerHit()
{
    double timeleft = CalcTimeLeft(MotionStart);
    uint32_t delay = 1000 * 30;   // inactive timer setting to maintain roof status lights
    if (!isConnected())
        return; //  No need to reset timer if we are not connected anymore

    if (isSimulation())
    {
        if (timeleft -5 <= 0)            // Use timeout approaching to set faux switch indicator
        {
            if (DomeMotionS[DOME_CW].s == ISS_ON)              // Opening
            {
                simRoofOpen = true;
                simRoofClosed = false;
            }
            else if (DomeMotionS[DOME_CCW].s == ISS_ON)        // Closing
            {
                simRoofClosed = true;
                simRoofOpen = false;
            }
        }
    }

    updateRoofStatus();

    if (DomeMotionSP.s == IPS_BUSY)
    {
        // Abort called stop movement.
        if (MotionRequest < 0)
        {
            DEBUG(INDI::Logger::DBG_WARNING,"Roof motion is stopped");
            setDomeState(DOME_IDLE);
        }
        else
        {
            // Roll off is opening
            if (DomeMotionS[DOME_CW].s == ISS_ON)
            {
                if (fullOpenLimitSwitch == ISS_ON)
                {
                    DEBUG(INDI::Logger::DBG_WARNING,"Roof is open");
                    SetParked(false);
                }
                // See if time to open has expired.
                else if (timeleft <= 0)
                {
                    LOG_WARN("Time allowed for opening the roof has expired?");
                    setDomeState(DOME_IDLE);
                    roofOpening = false;
                }
                else
                {
                    delay = 1000;           // opening active
                }
            }
            // Roll Off is closing
            else if (DomeMotionS[DOME_CCW].s == ISS_ON)
            {
                if (fullClosedLimitSwitch == ISS_ON)
                {
                    DEBUG(INDI::Logger::DBG_WARNING,"Roof is closed");
                    SetParked(true);
                }
                // See if time to open has expired.
                else if (timeleft <= 0)
                {
                    LOG_WARN("Time allowed for closing the roof has expired?");
                    setDomeState(DOME_IDLE);
                    roofClosing = false;
                }
                else
                {
                    delay = 1000;           // closing active
                }
            }
        }
    }

    // Even when no roof movement requested, will come through occasionally. Use timer to update roof status
    // in case roof has been operated externally by a remote control, locks applied...
    SetTimer(delay);
}

float RollOffIno::CalcTimeLeft(timeval start)
{
    double timesince;
    double timeleft;
    struct timeval now { 0, 0 };
    gettimeofday(&now, nullptr);

    timesince =
            (double)(now.tv_sec * 1000.0 + now.tv_usec / 1000) - (double)(start.tv_sec * 1000.0 + start.tv_usec / 1000);
    timesince = timesince / 1000;
    timeleft  = MotionRequest - timesince;
    return timeleft;
}

bool RollOffIno::saveConfigItems(FILE *fp)
{
    return INDI::Dome::saveConfigItems(fp);
}

/*
 * Direction: DOME_CW Clockwise = Open; DOME-CCW Counter clockwise = Close
 * Operation: MOTION_START, | MOTION_STOP
 */
IPState RollOffIno::Move(DomeDirection dir, DomeMotionCommand operation)
{
    updateRoofStatus();
    if (operation == MOTION_START)
    {
        if (roofLockedSwitch)
        {
            LOG_WARN("Roof is externally locked, no movement possible");
            return IPS_ALERT;
        }
        if (roofOpening)
        {
            LOG_WARN("Roof is in process of opening, wait for completion or abort current operation");
            return IPS_OK;
        }
        if (roofClosing)
        {
            LOG_WARN("Roof is in process of closing, wait for completion or abort current operation");
            return IPS_OK;
        }

        // Open Roof
        // DOME_CW --> OPEN. If we are asked to "open" while we are fully opened as the
        // limit switch indicates, then we simply return false.
        if (dir == DOME_CW)
        {
            if (fullOpenLimitSwitch == ISS_ON)
            {
                LOG_WARN("DOME_CW directive received but roof is already fully opened");
                return IPS_ALERT;
            }
            else if (getWeatherState() == IPS_ALERT)
            {
                LOG_WARN("Weather conditions are in the danger zone. Cannot open roof");
                return IPS_ALERT;
            }

            // Initiate action
            if (roofOpen())
            {
                roofOpening = true;
                roofClosing = false;
                LOG_INFO("Roof is opening...");
            }
            else
            {
                LOG_WARN("Failed to operate controller to open roof");
                return IPS_ALERT;
            }
        }

        // Close Roof
        else if (dir == DOME_CCW)
        {
            if (fullClosedLimitSwitch == ISS_ON)
            {
                LOG_WARN("DOME_CCW directive received but roof is already fully closed");
                return IPS_ALERT;
            }
            else if (INDI::Dome::isLocked())
            {
                DEBUG(INDI::Logger::DBG_WARNING,
                  "Cannot close dome when mount is locking. See: Telescope parkng policy, in options tab");
                return IPS_ALERT;
            }
            // Initiate action
            if (roofClose())
            {
                roofClosing = true;
                roofOpening = false;
                LOG_INFO("Roof is closing...");
            }
            else
            {
                LOG_WARN("Failed to operate controller to close roof");
                return IPS_ALERT;
            }
        }
        MotionRequest = ROLLOFF_DURATION;
        gettimeofday(&MotionStart, nullptr);
        SetTimer(1000);
        return IPS_BUSY;
    }
     return    IPS_ALERT;
}
/*
 * Close Roof
 *
 */
IPState RollOffIno::Park()
{
    IPState rc = INDI::Dome::Move(DOME_CCW, MOTION_START);

    if (rc == IPS_BUSY)
    {
        LOG_INFO("RollOff ino is parking...");
        return IPS_BUSY;
    }
    else
        return IPS_ALERT;
}

/*
 * Open Roof
 *
 */
IPState RollOffIno::UnPark()
{
    IPState rc = INDI::Dome::Move(DOME_CW, MOTION_START);
    if (rc == IPS_BUSY)
    {
        LOG_INFO("RollOff ino is unparking...");
        return IPS_BUSY;
    }
    else
        return IPS_ALERT;
}

/*
 * Abort motion
 */
bool RollOffIno::Abort()
{
    bool lockState;
    bool openState;
    bool closeState;

    updateRoofStatus();
    lockState = (roofLockedSwitch == ISS_ON);
    openState = (fullOpenLimitSwitch == ISS_ON);
    closeState = (fullClosedLimitSwitch == ISS_ON);

    if (lockState)
    {
        LOG_WARN("Roof is externally locked, no action taken on abort request");
        return true;
    }

    if (closeState && DomeMotionSP.s != IPS_BUSY)
    {
        LOG_WARN("Roof appears to be closed and stationary, no action taken on abort request");
        return true;
    }
    else if (openState && DomeMotionSP.s != IPS_BUSY)
    {
        LOG_WARN("Roof appears to be open and stationary, no action taken on abort request");
        return true;
    }
    else if (DomeMotionSP.s != IPS_BUSY)
    {
        LOG_WARN("Roof appears to be partially open and stationary, no action taken on abort request");
    }
    else if (DomeMotionSP.s == IPS_BUSY)
    {
        if (DomeMotionS[DOME_CW].s == ISS_ON)
        {
            LOG_WARN("Abort roof action requested. The roof appears to be opening, on the next move request it will close");
        }
        else if (DomeMotionS[DOME_CCW].s == ISS_ON)
        {
            LOG_WARN("Abort roof action requested. The roof appears to be closing, on the next move request it will open");
        }
        roofMoving = false;
        roofClosing = false;
        roofOpening = false;
        MotionRequest = -1;
        roofAbort();
    }

    // If both limit switches are off, then we're neither parked nor unparked.
    if (fullOpenLimitSwitch == ISS_OFF && fullClosedLimitSwitch == ISS_OFF)
    {
        IUResetSwitch(&ParkSP);
        ParkSP.s = IPS_IDLE;
        IDSetSwitch(&ParkSP, nullptr);
    }
    return true;
}

bool RollOffIno::getFullOpenedLimitSwitch(bool* switchState)
{
    if (isSimulation())
    {
        if (simRoofOpen)
        {
            fullOpenLimitSwitch = ISS_ON;
            *switchState = true;
        }
        else
        {
            fullOpenLimitSwitch = ISS_OFF;
            *switchState = false;
        }
        return true;
    }

    if (readRoofSwitch(ROOF_OPENED_SWITCH, switchState))
    {
        if (*switchState)
            fullOpenLimitSwitch = ISS_ON;
        else
            fullOpenLimitSwitch = ISS_OFF;
        return true;
    }
    else
    {
        LOG_WARN("Unable to obtain from the controller whether or not the roof is opened");
        return false;
    }
}

bool RollOffIno::getFullClosedLimitSwitch(bool* switchState)
{
    if (isSimulation())
    {
        if (simRoofClosed)
        {
            fullClosedLimitSwitch = ISS_ON;
            *switchState = true;
        }
        else
        {
            fullClosedLimitSwitch = ISS_OFF;
            *switchState = false;
        }
        return true;
    }

    if (readRoofSwitch(ROOF_CLOSED_SWITCH, switchState))
    {
        if (*switchState)
            fullClosedLimitSwitch = ISS_ON;
        else
            fullClosedLimitSwitch = ISS_OFF;
        return true;
    }
    else
    {
        LOG_WARN("Unable to obtain from the controller whether or not the roof is closed");
        return false;
    }
}

bool RollOffIno::getRoofLockedSwitch(bool* switchState)
{
    // If there is no lock switch, return success with status false
    if (isSimulation())
    {
        roofLockedSwitch = ISS_OFF;
        *switchState = false;                     // Not locked
        return true;
    }
    if (readRoofSwitch(ROOF_LOCKED_SWITCH, switchState))
    {
        if (*switchState)
            roofLockedSwitch = ISS_ON;
        else
            roofLockedSwitch = ISS_OFF;
        return true;
    }
    else
    {
        LOG_WARN("Unable to obtain from the controller whether or not the roof is externally locked");
        return false;
    }
}

    bool RollOffIno::getRoofAuxSwitch(bool* switchState)
    {
        // If there is no lock switch, return success with status false
        if (isSimulation())
        {
            if (AuxS[AUX_ENABLE].s == ISS_OFF)
            {
                roofAuxSwitch = ISS_OFF;
                *switchState = false;
                return true;
            }
            else
            {
                roofAuxSwitch = ISS_ON;
                *switchState = true;
                return true;
            }
        }

        if (readRoofSwitch(ROOF_AUX_SWITCH, switchState))
        {
            if (*switchState)
                roofAuxSwitch = ISS_ON;
            else
                roofAuxSwitch = ISS_OFF;
            return true;
        }
        else
        {
            LOG_WARN("Unable to obtain from the controller whether or not the obs Aux switch is being used");
            return false;
        }
    }


/*
 * -------------------------------------------------------------------------------------------
 *
 */
bool RollOffIno::roofOpen()
{
    if (isSimulation())
    {
        return true;
    }
    return pushRoofButton(ROOF_OPEN_RELAY);
}

bool RollOffIno::roofClose()
{
    if (isSimulation())
    {
        return true;
    }
    return pushRoofButton(ROOF_CLOSE_RELAY);
}

bool RollOffIno::roofAbort()
{
    if (isSimulation())
    {
        return true;
    }
    return pushRoofButton(ROOF_ABORT_RELAY);
}

bool RollOffIno::roofAux()
{
    if (isSimulation())
    {
        return false;
    }
    return pushRoofButton(ROOF_AUX_RELAY);
}

/*
 * If unable to determine switch state due to errors, return false.
 * If no errors return true. Return in result true if switch and false if switch off.
 */
bool RollOffIno::readRoofSwitch(const char* roofSwitchId, bool *result)
{
    char readBuffer[MAXINOBUF];
    char writeBuffer[MAXINOLINE];
    bool status;

    if (!contactEstablished)
    {
        LOG_WARN("No contact with the roof controller has been established");
        return false;
    }
    if (roofSwitchId == 0)
        return false;
    memset(writeBuffer, 0, sizeof(writeBuffer));
    strcpy(writeBuffer, "(GET:");
    strcat(writeBuffer, roofSwitchId);
    strcat(writeBuffer, ":0)");
    if (!writeIno(writeBuffer))
        return false;
    memset(readBuffer, 0, sizeof(readBuffer));
    if (!readIno(readBuffer))
        return false;
    status = evaluateResponse(readBuffer, result);
    return status;
}

/*
 * See if the controller is running
 */
bool RollOffIno::initialContact(void)
{
    char readBuffer[MAXINOBUF];
    bool result = false;
    contactEstablished = false;
    if (writeIno("(CON:0:0)"))
    {
        memset(readBuffer, 0, sizeof(readBuffer));
        if (readIno(readBuffer))
        {
            contactEstablished = evaluateResponse(readBuffer, &result);
            return contactEstablished;
        }
    }
    LOG_ERROR("Unable to contact the roof controller");
    return false;
}

/*
 * Whether roof is moving or stopped in any position along with the nature of the button requested will
 * determine the effect on the roof. This could mean stopping, or starting in a reversed direction.
 */
bool RollOffIno::pushRoofButton(const char* button)
{
    char readBuffer[MAXINOBUF];
    char writeBuffer[MAXINOBUF];
    bool status;
    bool switchState = false;

    if (!contactEstablished)
    {
        LOG_WARN("No contact with the roof controller has been established");
        return false;
    }
    status = getRoofLockedSwitch(&switchState);        // In case it has been locked since the driver connected
    if (status && !switchState)
    {
        memset(writeBuffer, 0, sizeof(writeBuffer));
        strcpy(writeBuffer, "(SET:");
        strcat(writeBuffer, button);
        strcat(writeBuffer, ":ON)");
        if (!writeIno(writeBuffer))             // Push identified button & get response
            return false;
        msSleep(ROR_D_PRESS);
        memset(readBuffer, 0, sizeof(readBuffer));
        return readIno(readBuffer);
    }
    else
    {
        LOG_WARN("Roof external lock state prevents roof movement");
        return false;
    }
}

/*
 * if ACK return true and set result true|false indicating if switch is on
 *
 */
 bool RollOffIno::evaluateResponse(char* buff, bool* result)
{
    char inoCmd[MAXINOCMD+1];
    char inoTarget[MAXINOTARGET+1];
    char inoVal[MAXINOVAL+1];

    *result = false;
    strcpy(inoCmd, strtok(buff, "(:"));
    strcpy(inoTarget, strtok(nullptr,":"));
    strcpy(inoVal, strtok(nullptr,")"));
    LOGF_DEBUG("Returned from roof controller: Cmd: %s, Target: %s, Value: %s", inoCmd, inoTarget, inoVal);
    if ((strcmp(inoCmd,"NAK")) == 0)
    {
        LOGF_WARN("Negative response from roof controller error: %s", inoVal);
        return false;
    }
    *result = (strcmp(inoVal, "ON") == 0);
    return true;
}

bool RollOffIno::readIno(char* retBuf)
{
    bool stop = false;
    bool start_found = false;
    int status;
    int retCount = 0;
    int totalCount = 0;
    char* bufPtr = retBuf;
    char errMsg[MAXINOERR];

    while (!stop)
    {
        bufPtr = bufPtr + retCount;
        status = tty_read(PortFD, bufPtr, 1, MAXINOWAIT, &retCount);
        if (status != TTY_OK)
        {
            tty_error_msg(status, errMsg, MAXINOERR);
            LOGF_ERROR("roof control connection error: %s", errMsg);
            return false;
        }
        if (retCount > 0)
        {
            if (*bufPtr == 0X28)             // '('   Start found
                start_found = true;
            if (!start_found)
                retCount = 0;
            totalCount += retCount;
            if ((*bufPtr == 0X29) || (totalCount >= MAXINOBUF-2))    // ')'   End found
            {
                *(++bufPtr) = 0;
                stop = true;
            }
        }
    }
    return true;
}

bool RollOffIno::writeIno(const char* msg)
{
    int retMsgLen = 0;
    int status;
    char errMsg[MAXINOERR];

    if (strlen(msg) >= MAXINOLINE)
    {
        LOG_ERROR("Roof controller command message too long");
        return false;
    }
    LOGF_DEBUG("Sent to roof controller: %s", msg);
    tcflush(PortFD, TCIOFLUSH);
    status = tty_write_string(PortFD, msg, &retMsgLen);
    if (status != TTY_OK)
    {
        tty_error_msg(status, errMsg, MAXINOERR);
        LOGF_ERROR("roof control connection error: %s", errMsg);
        return false;
    }
    return true;
}

void RollOffIno::msSleep (int mSec)
{
    struct timespec req = {0,0};
    req.tv_sec = 0;
    req.tv_nsec = mSec * 1000000L;
    nanosleep(&req, (struct timespec *)nullptr);
}

